#include <cmath>
#include <fstream>
#include <functional>
#include <memory>
#include <thread>
#include "absl/flags/flag.h"
#include "absl/flags/parse.h"
#include "absl/time/clock.h"
#include "absl/time/time.h"
#include "glog/logging.h"

#include "sequence_compare/hamming_distance.h"

ABSL_FLAG(std::string, input, "", "Input file name.");
ABSL_FLAG(std::string, output, "", "Output file name.");
ABSL_FLAG(int, num_threads, 1, "Number of threads.");
ABSL_FLAG(int, window_size, 3000, "Window size.");
ABSL_FLAG(int, step_size, 1000, "Step window size");
ABSL_FLAG(
    bool, append_sites_compared, false,
    "Whether to append number of sites compared to the output of each pair.");

namespace {

// Write key and value to data, clear the contents of key and value.
void StoreKVPair(std::unique_ptr<std::string>& key,
                 std::unique_ptr<std::string>& value,
                 std::vector<std::pair<std::string, std::string>>& data) {
  const std::string kValidSymbols = "ATCGatcg";
  CHECK(key);
  CHECK(value);
  CHECK(key->empty() == value->empty());

  // Store previous KV pair.
  if (!key->empty()) {
    std::replace_if(value->begin(), value->end(),
                    [&kValidSymbols](const char c) {
                      return kValidSymbols.find(c) == std::string::npos;
                    },
                    '-');

    data.emplace_back(std::move(*key), std::move(*value));
    key.reset(new std::string());
    value.reset(new std::string());
  }
}

void ParseFASTA(const std::string& file_name,
                std::vector<std::pair<std::string, std::string>>& data) {
  std::ifstream fin(file_name);
  CHECK(fin.good());

  std::unique_ptr<std::string> key, value;

  std::string line;
  key.reset(new std::string());
  value.reset(new std::string());

  while (fin.good() && !fin.eof()) {
    std::getline(fin, line);

    if (line.front() == '>') {
      StoreKVPair(key, value, data);

      CHECK_GT(line.size(), 1u);
      line.erase(line.begin());
      *key = line;
    } else {
      *value += line;
    }
  }
  StoreKVPair(key, value, data);

  CHECK(!data.empty());
  for (const std::pair<std::string, std::string>& kv : data) {
    CHECK_EQ(kv.second.size(), data.front().second.size())
        << kv.second.size() << " is not equal to " << data.front().second.size()
        << ".";
  }

  LOG(INFO) << "Read " << data.size() << " entries from " << file_name
            << ". Each sequence has length " << data.front().second.size();
}

inline void WriteData(std::ofstream& fout,
                      const hamming_distance::dist_data* hamming_dist) {
  for (const std::pair<std::pair<absl::string_view, absl::string_view>,
                       std::vector<int>>& datum : *hamming_dist) {
    fout << datum.first.first << "," << datum.first.second;
    for (int idx : datum.second) {
      fout << "," << idx;
    }
    fout << std::endl;
  }
}

void WriteData(const std::string& file_name,
               const std::vector<std::unique_ptr<hamming_distance::dist_data>>&
                   hamming_dists) {
  std::ofstream fout(file_name);
  CHECK(fout.good());
  for (const std::unique_ptr<hamming_distance::dist_data>& hamming_dist :
       hamming_dists) {
    WriteData(fout, hamming_dist.get());
  }
}
}  // namespace

int main(int argc, char** argv) {
  FLAGS_alsologtostderr = true;
  absl::ParseCommandLine(argc, argv);

  CHECK_GE(absl::GetFlag(FLAGS_num_threads), 1);
  const int kParallel = absl::GetFlag(FLAGS_num_threads);

  std::vector<std::pair<std::string, std::string>> data;
  std::vector<std::unique_ptr<hamming_distance::dist_data>> hamming_dists;
  std::vector<std::thread> threads;

  ParseFASTA(absl::GetFlag(FLAGS_input), data);

  absl::Time start_t = absl::Now();
  for (int i = 0; i < kParallel; ++i) {
    int start = data.size() * sqrt(static_cast<double>(i) / kParallel);
    int end = (i + 1) == kParallel
                  ? data.size()
                  : data.size() * sqrt(static_cast<double>(i + 1) / kParallel);
    hamming_dists.emplace_back(new hamming_distance::dist_data);
    threads.emplace_back(
        hamming_distance::CompareWorker, std::ref(data), start, end,
        absl::GetFlag(FLAGS_window_size), absl::GetFlag(FLAGS_step_size),
        absl::GetFlag(FLAGS_append_sites_compared), hamming_dists.back().get());
    LOG(INFO) << "Launched worker" << start << ":" << end << ".";
  }
  for (int i = 0; i < kParallel; ++i) {
    threads[i].join();
  }
  LOG(INFO) << "Thread process time: "
            << absl::ToDoubleSeconds(absl::Now() - start_t) << " seconds.";

  WriteData(absl::GetFlag(FLAGS_output), hamming_dists);

  return 0;
}
